from typing import Callable, Optional, Dict, Any, Union, Mapping, Iterable
from types import FrameType, MethodType, FunctionType, BuiltinFunctionType
from collections import OrderedDict
import torch
from torch import nn
from torch.nn.parameter import UninitializedTensorMixin
import pyreflex
from pyreflex import typeattr, hidden, HiddenMeta
from pyreflex.pybase import framelocals
from packaging import version


def _addindent(s_: str, numSpaces):
    s = s_.split('\n')
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


class _BufferingModule:
    # This class instance will replace the `nn.Module` instance `self` in the `register_parameter` when registering a buffer,
    # while the `self` in the `__setattr__` will keep same.
    class FakeDict:
        def __setitem__(self, _, __): pass
    def __init__(self):
        self._parameters = _BufferingModule.FakeDict()


if version.parse(torch.__version__) >= version.parse('2.0.0'):
    class _BufferingFakeDict:
        # This class instance will replace the global variable `_global_parameter_registration_hooks` in the `register_parameter`
        # when registering a buffer, and it returns a iterator when `values()` is called. When iterating, the original
        # `_global_parameter_registration_hooks` will be restored.
        class Iterator:
            def __init__(self, frame, global_parameter_registration_hooks):
                self.frame = frame
                self.global_parameter_registration_hooks = global_parameter_registration_hooks
                
            def __iter__(self):
                return self
        
            def __next__(self):
                # Retore the original `_global_parameter_registration_hooks`.
                self.frame.f_globals['_global_parameter_registration_hooks'] = self.global_parameter_registration_hooks
                raise StopIteration
        
        def __init__(self, frame, global_parameter_registration_hooks):
            self.frame = frame
            self.global_parameter_registration_hooks = global_parameter_registration_hooks
        
        def values(self):
            return _BufferingFakeDict.Iterator(self.frame, self.global_parameter_registration_hooks)
    
    def _hook_global_parameter_registration(last_frame: FrameType):
        global_parameter_registration_hooks = last_frame.f_globals['_global_parameter_registration_hooks']
        last_frame.f_globals['_global_parameter_registration_hooks'] = _BufferingFakeDict(last_frame, global_parameter_registration_hooks)
else:
    def _hook_global_parameter_registration(_): pass


def bufferingmethod(funcobj):
    # Only to tell `typeattr` that only the "buffering method" is required to be called.
    funcobj.__isbufferingmethod__ = True
    return funcobj


def _is_buffering_method(item):
    # The condition itself for a `typeattr` instance, indicating that only the
    # "buffering method" is required to be called.
    return getattr(item, '__isbufferingmethod__', False)


class Buffering(metaclass=HiddenMeta):
    '''
    A class used to represent a content as a buffer, which can be used in the `__setattr__` of the `nn.Module` instance,
    or registered by the method `register_buffer` called by `nn.Module` instance.
    '''
    @bufferingmethod
    @hidden
    def register_to(self, module: nn.Module, name: str):
        module.register_buffer(name, typeattr(self, _is_buffering_method)[__class__.get_content.__code__.co_name](self), typeattr(self, _is_buffering_method)[__class__.get_persistent.__code__.co_name](self))
    
    @bufferingmethod
    @hidden
    def modify_generically(self, func: Callable[[Any], Any], *args, **kwargs): pass
    
    @bufferingmethod
    @hidden
    def copy_generically(self, func: Callable[[Any], Any], *args, **kwargs): pass
    
    @bufferingmethod
    @hidden
    @staticmethod
    def standard_type(): pass
    
    @bufferingmethod
    @hidden
    def get_apparent_type(self) -> Any: pass
    
    @bufferingmethod
    @hidden
    def set_apparent_type(self, value): pass
    
    @bufferingmethod
    @hidden
    def get_attr(self, name: str) -> Any: pass

    @bufferingmethod
    @hidden
    def set_attribute(self, name: str, value): pass
    
    @bufferingmethod
    @hidden
    def delete_attribute(self, name: str): pass
    
    @bufferingmethod
    @hidden
    def get_content(self) -> Any: pass
    
    @bufferingmethod
    @hidden
    def get_persistent(self) -> bool: pass
    
    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        return torch.Tensor.__torch_function__.__func__(cls, func, types, args, kwargs)
    
    def __getattr__(self, name: str):
        if name == 'grad_fn':
            # The attibute `grad_fn` is called by the `register_parameter` method, so it can be replaced by this getter to
            # insert any operations, which ignores the storage in `register_parameter` as a parameter, instaed, `register_buffer`
            # will be called here.
            i = pyreflex.overriding_depth()
            first_frame = pyreflex.get_frame_at(i)
            second_frame = pyreflex.get_frame_at(i + 1)
            if (first_frame and first_frame.f_code.co_name == 'register_parameter') and (second_frame and second_frame.f_code.co_name == '__setattr__'):
                # Only if the last frame function is `register_parameter` and the last second frame fuction is `__setattr__`, 
                # which means it is in the parameterized hook process, this buffer storage will be called, else `grad_fn`
                # will be gotten normally.
                last_frame = first_frame
                module_self: nn.Module = pyreflex.self_from_frame(last_frame)
                if module_self is not None and isinstance(module_self, nn.Module):
                    co_varnames = last_frame.f_code.co_varnames
                    if len(co_varnames) > 1:
                        last_self_name = co_varnames[0]
                        last_name = last_frame.f_locals[co_varnames[1]]
                        
                        framelocals(last_frame)[last_self_name] = _BufferingModule()
                        _hook_global_parameter_registration(last_frame)
                        
                        typeattr(self, _is_buffering_method)[__class__.register_to.__code__.co_name](self, module_self, last_name)
                        return False
        attr = typeattr(self, _is_buffering_method)[__class__.get_attr.__code__.co_name](self, name)
        return attr
    
    def __setattr__(self, name: str, value):
        if name == '__class__':
            # Only the attribute `__class__` is required to be set using the property setter.
            super().__setattr__(name, value)
        else:
            # All other attributes should be set in the stored content.
            typeattr(self, _is_buffering_method)[__class__.set_attribute.__code__.co_name](self, name, value)
    
    def __delattr__(self, name: str):
        typeattr(self, _is_buffering_method)[__class__.delete_attribute.__code__.co_name](self, name)
    
    def __repr__(self) -> str:
        return f"<unregistered buffer of {typeattr(self, _is_buffering_method)[__class__.get_content.__code__.co_name](self).__repr__()}>"
    
    @property
    def __class__(self):
        apparent_type = None
        standard_type = typeattr(self, _is_buffering_method)[__class__.standard_type.__code__.co_name]()
        i = pyreflex.overriding_depth()
        while pyreflex.get_frame_at(i).f_code.co_name == '__getattribute__':
            i += 1
        first_frame = pyreflex.get_frame_at(i)
        second_frame = pyreflex.get_frame_at(i + 1)
        third_frame = pyreflex.get_frame_at(i + 2)
        if second_frame and (second_frame.f_code.co_name == '__setattr__' or 
            (third_frame and second_frame.f_code.co_name == 'register_parameter' and third_frame.f_code.co_name == '__setattr__')):
            # Consider the current frame function is `__instancecheck__`, if the last frame function is `__setattr__`,
            # or the last frame function is `register_parameter` and the last second frame function is `__setattr__`,
            # that means it is in the parameterized hook process.
            frame = second_frame
            module_self: nn.Module = pyreflex.self_from_frame(frame)
            apparent_type = typeattr(self, _is_buffering_method)[__class__.get_apparent_type.__code__.co_name](self)
            if module_self is not None and isinstance(module_self, nn.Module) and issubclass(apparent_type, standard_type):
                # When the `__class__` of the stored content is `standard_type` or its subclass, returning the `nn.Parameter`
                # to hook the `__instancecheck__`.
                return nn.Parameter
        elif first_frame and first_frame.f_code.co_name == 'register_buffer':
            # When the buffer is passed to the `register_buffer` method, the stored content and its persistent status will
            # be extracted into the `register_buffer` method. (Strangely, the `__instancecheck__` is not called, so the
            # stack index is less.)
            frame = first_frame
            module_self: nn.Module = pyreflex.self_from_frame(frame)
            if module_self is not None and isinstance(module_self, nn.Module):
                co_varnames = frame.f_code.co_varnames
                if len(co_varnames) > 3:
                    from pyreflex import get_instruction_at
                    tensor_name = co_varnames[2]
                    persistent_name = co_varnames[3]
                    lasti = frame.f_back.f_lasti
                    call_instruction = get_instruction_at(frame.f_back.f_code, lasti)
                    frame_locals = framelocals(frame)
                    if call_instruction.argval is not None and call_instruction.argval <= 2:
                        # Only when explicitly passing the `persistent` argument to the method `register_buffer`, the passed
                        # one will be applied, otherwise the `persistent` attribute in the buffer instance will be exploited.
                        # Meanwhile, using the form of "module.register_buffer(*args, **kwargs)" will only ignore `persistent`
                        # argument of the method, only keep the buffer's setting.
                        frame_locals[persistent_name] = typeattr(self, _is_buffering_method)[__class__.get_persistent.__code__.co_name](self)
                    frame_locals[tensor_name] =  typeattr(self, _is_buffering_method)[__class__.get_content.__code__.co_name](self)
                    
                    apparent_type = typeattr(self, _is_buffering_method)[__class__.get_apparent_type.__code__.co_name](self)
                    if issubclass(apparent_type, standard_type):
                        return torch.Tensor
        elif second_frame.f_code.co_name == '_load_from_state_dict' and first_frame.f_code.co_name == 'is_lazy' and issubclass(type(self), BufferObject):
            return UninitializedTensorMixin
        
        return apparent_type if apparent_type else typeattr(self, _is_buffering_method)[__class__.get_apparent_type.__code__.co_name](self)
    
    @__class__.setter
    def __class__(self, value):
        # The `__class__` should be set through the stored content.
        typeattr(self, _is_buffering_method)[__class__.set_apparent_type.__code__.co_name](self, value)
    
    def __copy__(self):
        from copy import copy
        return typeattr(self, _is_buffering_method)[__class__.copy_generically.__code__.co_name](self, copy)
    
    def __deepcopy__(self, memodict={}):
        from copy import deepcopy
        return typeattr(self, _is_buffering_method)[__class__.copy_generically.__code__.co_name](self, deepcopy, memodict)


class Buffer(Buffering):
    def __init__(self, tensor: Optional[torch.Tensor], persistent: bool = True):
        object.__setattr__(self, '__content', tensor)
        object.__setattr__(self, '__persistent', persistent)
    
    @bufferingmethod
    @hidden
    def modify_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        content = object.__getattribute__(self, '__content')
        result = func(content, *args, **kwargs)
        if isinstance(result, torch.Tensor) or issubclass(type(result), Buffering):
            object.__setattr__(self, '__content', result)
        else:
            return result
        return self
    
    @bufferingmethod
    @hidden
    def copy_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        content = object.__getattribute__(self, '__content')
        result = func(content, *args, **kwargs)
        if isinstance(result, torch.Tensor) or issubclass(type(result), Buffering):
            instance = type(self).__new__(type(self))
            object.__setattr__(instance, '__content', result)
        else:
            return result
        object.__setattr__(instance, '__persistent', object.__getattribute__(self, '__persistent'))
        return instance
    
    @bufferingmethod
    @hidden
    @staticmethod
    def standard_type():
        return torch.Tensor
    
    @bufferingmethod
    @hidden
    def get_apparent_type(self) -> Any:
        return object.__getattribute__(self, '__content').__class__
    
    @bufferingmethod
    @hidden
    def set_apparent_type(self, value):
        object.__getattribute__(self, '__content').__class__ = value
    
    @bufferingmethod
    @hidden
    def get_attr(self, name: str) -> Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            content = object.__getattribute__(self, '__content')
            attr = content.__getattribute__(name)
            # If the instance is a builtin function of `torch.Tensor`, then wrap it through `modify_generically` when the function name ends with "_",
            # otherwise through `copy_generically`.
            def when_function(func):
                if name.endswith('_'):
                    def required_function(instance, *args, **kwargs):
                        return typeattr(instance, _is_buffering_method)[__class__.modify_generically.__code__.co_name](instance, func, *args, **kwargs)
                else:
                    def required_function(instance, *args, **kwargs):
                        return typeattr(instance, _is_buffering_method)[__class__.copy_generically.__code__.co_name](instance, func, *args, **kwargs)
                required_function.__name__ = name
                required_function.__qualname__ = name
                return MethodType(required_function, self)
            if type(attr) is BuiltinFunctionType:
                func = attr
                attr = when_function(lambda _, *args, **kwargs: func(*args, **kwargs))
            elif type(attr) is MethodType:
                attr = when_function(attr.__func__)
            elif type(attr) is FunctionType:
                attr = when_function(attr)
            return attr

    @bufferingmethod
    @hidden
    def set_attribute(self, name: str, value):
        object.__getattribute__(self, '__content').__setattr__(name, value)
    
    @bufferingmethod
    @hidden
    def delete_attribute(self, name: str):
        object.__getattribute__(self, '__content').__delattr__(name)
    
    @bufferingmethod
    @hidden
    def get_content(self):
        return object.__getattribute__(self, '__content')
    
    @bufferingmethod
    @hidden
    def get_persistent(self):
        return object.__getattribute__(self, '__persistent')

    # def __copy__(self):
    #     from copy import copy
    #     return Buffer(copy(object.__getattribute__(self, '__content')), object.__getattribute__(self, '__persistent'))
    
    # def __deepcopy__(self, memodict={}):
    #     from copy import deepcopy
    #     return Buffer(deepcopy(object.__getattribute__(self, '__content'), memodict), object.__getattribute__(self, '__persistent'))
    
    def __del__(self): object.__delattr__(self, '__content')
    def __cmp__(self, other): return object.__getattribute__(self, '__content').__cmp__(other)
    def __eq__(self, other): return object.__getattribute__(self, '__content') == other
    def __ne__(self, other): return object.__getattribute__(self, '__content') != other
    def __lt__(self, other): return object.__getattribute__(self, '__content') < other
    def __gt__(self, other): return object.__getattribute__(self, '__content') > other
    def __le__(self, other): return object.__getattribute__(self, '__content') <= other
    def __ge__(self, other): return object.__getattribute__(self, '__content') >= other
    def __pos__(self): return +object.__getattribute__(self, '__content')
    def __neg__(self): return -object.__getattribute__(self, '__content')
    def __abs__(self): return abs(object.__getattribute__(self, '__content'))
    def __invert__(self): return ~object.__getattribute__(self, '__content')
    def __round__(self, n): return round(object.__getattribute__(self, '__content'), n)
    def __floor__(self): return object.__getattribute__(self, '__content').__floor__()
    def __ceil__(self): return object.__getattribute__(self, '__content').__ceil__()
    def __trunc__(self): return object.__getattribute__(self, '__content').__trunc__()
    def __add__(self, other): return object.__getattribute__(self, '__content') + other
    def __sub__(self, other): return object.__getattribute__(self, '__content') - other
    def __mul__(self, other): return object.__getattribute__(self, '__content') * other
    def __floordiv__(self, other): return object.__getattribute__(self, '__content') // other
    def __div__(self, other): return object.__getattribute__(self, '__content') / other
    def __truediv__(self, other): return object.__getattribute__(self, '__content').__truediv__(other)
    def __mod__(self, other): return object.__getattribute__(self, '__content') % other
    def __divmod__(self, other): return divmod(object.__getattribute__(self, '__content'), other)
    def __pow__(self, other): return object.__getattribute__(self, '__content') ** other
    def __lshift__(self, other): return object.__getattribute__(self, '__content') << other
    def __rshift__(self, other): return object.__getattribute__(self, '__content') >> other
    def __and__(self, other): return object.__getattribute__(self, '__content') & other
    def __or__(self, other): return object.__getattribute__(self, '__content') | other
    def __xor__(self, other): return object.__getattribute__(self, '__content') ^ other
    def __radd__(self, other): return other + object.__getattribute__(self, '__content')
    def __rsub__(self, other): return other - object.__getattribute__(self, '__content')
    def __rmul__(self, other): return other * object.__getattribute__(self, '__content')
    def __rfloordiv__(self, other): return other // object.__getattribute__(self, '__content')
    def __rdiv__(self, other): return other / object.__getattribute__(self, '__content')
    def __rtruediv__(self, other): return object.__getattribute__(self, '__content').__rtruediv__(other)
    def __rmod__(self, other): return other % object.__getattribute__(self, '__content')
    def __rdivmod__(self, other): return divmod(other, object.__getattribute__(self, '__content'))
    def __rpow__(self, other): return other ** object.__getattribute__(self, '__content')
    def __rlshift__(self, other): return other << object.__getattribute__(self, '__content')
    def __rrshift__(self, other): return other >> object.__getattribute__(self, '__content')
    def __rand__(self, other): return other & object.__getattribute__(self, '__content')
    def __ror__(self, other): return other | object.__getattribute__(self, '__content')
    def __rxor__(self, other): return other ^ object.__getattribute__(self, '__content')
    def __iadd__(self, other): object.__getattribute__(self, '__content').__iadd__(other)
    def __isub__(self, other): object.__getattribute__(self, '__content').__isub__(other)
    def __imul__(self, other): object.__getattribute__(self, '__content').__imul__(other)
    def __ifloordiv__(self, other): object.__getattribute__(self, '__content').__ifloordiv__(other)
    def __idiv__(self, other): object.__getattribute__(self, '__content').__idiv__(other)
    def __itruediv__(self, other): object.__getattribute__(self, '__content').__itruediv__(other)
    def __imod__(self, other): object.__getattribute__(self, '__content').__imod__(other)
    def __ipow__(self, other): object.__getattribute__(self, '__content').__ipow__(other)
    def __ilshift__(self, other): object.__getattribute__(self, '__content').__ilshift__(other)
    def __irshift__(self, other): object.__getattribute__(self, '__content').__irshift__(other)
    def __iand__(self, other): object.__getattribute__(self, '__content').__iand__(other)
    def __ior__(self, other): object.__getattribute__(self, '__content').__ior__(other)
    def __ixor__(self, other): object.__getattribute__(self, '__content').__ixor__(other)
    def __int__(self): return object.__getattribute__(self, '__content').__int__()
    def __long__(self): return object.__getattribute__(self, '__content').__long__()
    def __float__(self): return object.__getattribute__(self, '__content').__float__()
    def __complex__(self): return object.__getattribute__(self, '__content').__complex__()
    def __oct__(self): return object.__getattribute__(self, '__content').__oct__()
    def __hex__(self): return object.__getattribute__(self, '__content').__hex__()
    def __index__(self): return object.__getattribute__(self, '__content').__index__()
    def __trunc__(self): return object.__getattribute__(self, '__content').__trunc__()
    def __coerce__(self, other): return object.__getattribute__(self, '__content').__coerce__(other)
    def __str__(self): return object.__getattribute__(self, '__content').__str__()
    def __unicode__(self): return object.__getattribute__(self, '__content').__unicode__()
    def __format__(self, formatstr): return object.__getattribute__(self, '__content').__format__(formatstr)
    def __hash__(self): return object.__getattribute__(self, '__content').__hash__()
    def __nonzero__(self): return object.__getattribute__(self, '__content').__nonzero__()
    def __dir__(self): return object.__getattribute__(self, '__content').__dir__()
    def __sizeof__(self): return object.__getattribute__(self, '__content').__sizeof__()
    def __len__(self): return len(object.__getattribute__(self, '__content'))
    def __getitem__(self, key): return object.__getattribute__(self, '__content')[key]
    def __setitem__(self, key, value): object.__getattribute__(self, '__content')[key] = value
    def __delitem__(self, key): object.__getattribute__(self, '__content').__delitem__(key)
    def __iter__(self): return iter(object.__getattribute__(self, '__content'))
    def __reversed__(self): return reversed(object.__getattribute__(self, '__content'))
    def __contains__(self, item): return item in object.__getattribute__(self, '__content')
    def __missing__(self, key): return object.__getattribute__(self, '__content').__missing__(key)
    def __call__(self, *args, **kwargs): return object.__getattribute__(self, '__content')(*args, **kwargs)
    def __enter__(self): return object.__getattribute__(self, '__content').__enter__()
    def __exit__(self, exception_type, exception_value, traceback): return object.__getattribute__(self, '__content').__exit__(exception_type, exception_value, traceback)
    def __get__(self, instance, owner): return object.__getattribute__(self, '__content').__get__(instance, owner)
    def __set__(self, instance, value): object.__getattribute__(self, '__content').__set__(instance, value)
    def __delete__(self, instance): return object.__getattribute__(self, '__content').__delete__(instance)


def _content_named_forward(content, *args, **kwargs):
    # Perfectly forward the called method name of the `BufferObject` instance.
    try:
        return content.__getattribute__(pyreflex.function_name(3))(*args, **kwargs)
    except AttributeError:
        # Sometimes, when the `_contents` dictionary in the `BufferObject` instance contains the
        # element that is not a `torch.Tensor` or something that cannot be forwarded the function
        # name that originally comes from the `torch.Tensor`, returning the element itself is the
        # best trade-off operation, this situation basically occurs when the type is `BufferDict`
        # or `BufferList`.
        return content


# def _content_named_forward_inplace(content, *args, **kwargs):
#     # Specifically, when the method name ends with "_", it means it is an inplace operation
#     try:
#         return content.__getattribute__(pyreflex.function_name(3)[:-1])(*args, **kwargs)
#     except AttributeError:
#         return content


def _named_forward_copy(instance, *args, **kwargs):
    return typeattr(instance, _is_buffering_method)[Buffering.copy_generically.__code__.co_name](instance, _content_named_forward, *args, **kwargs)


def _named_forward_modify(instance, *args, **kwargs):
    # Originally using `_content_named_forward_inplace`.
    return typeattr(instance, _is_buffering_method)[Buffering.modify_generically.__code__.co_name](instance, _content_named_forward, *args, **kwargs)


class BufferObject(Buffering):
    _contents: Dict[str, Union[torch.Tensor, Buffering, None]]
    
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        object.__setattr__(instance, '__class', cls)
        object.__setattr__(instance, '__persistent', True)
        object.__setattr__(instance, '_contents', cls.tensors())
        return instance
    
    @property
    def is_meta(self):
        return False
    
    @bufferingmethod
    @hidden
    def modify_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        contents = object.__getattribute__(self, '_contents')
        for name, content in contents.items():
            contents[name] = func(content, *args, **kwargs)
        return self
    
    @bufferingmethod
    @hidden
    def copy_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        instance = super().__new__(type(self))
        self_dict: dict = object.__getattribute__(self, '__dict__')
        for name, value in self_dict.items():
            type(self).__setattr__(instance, name, value)
        for name, content in object.__getattribute__(self, '_contents').items():
            object.__getattribute__(instance, '_contents')[name] = func(content, *args, **kwargs)
        return instance
    
    @bufferingmethod
    @hidden
    @staticmethod
    def tensors():
        return OrderedDict()
    
    @property
    def persistent(self):
        return object.__getattribute__(self, '__persistent')
    
    @persistent.setter
    def persistent(self, value):
        object.__setattr__(self, '__persistent', value)
    
    def with_persistent(self, value):
        self.persistent = value
        return self
    
    @bufferingmethod
    def register_to(self, module: nn.Module, name: str):
        module.register_buffer(name, self, object.__getattribute__(self, '__persistent'))
    
    @bufferingmethod
    @hidden
    @staticmethod
    def standard_type():
        return __class__
    
    @bufferingmethod
    @hidden
    def get_apparent_type(self) -> Any:
        return object.__getattribute__(self, '__class')
    
    @bufferingmethod
    @hidden
    def set_apparent_type(self, value):
        object.__setattr__(self, '__class', value)
    
    @bufferingmethod
    @hidden
    def get_attr(self, name: str) -> Any:
        attr = object.__getattribute__(self, '_contents').get(name)
        if attr is None:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
        else:
            return attr

    @bufferingmethod
    @hidden
    def set_attribute(self, name: str, value):
        object.__getattribute__(self, '_contents').pop(name, None)
        try:
            object.__delattr__(self, name)
        except AttributeError: ...
        if isinstance(value, torch.Tensor) or isinstance(value, Buffering):
            object.__getattribute__(self, '_contents')[name] = value
        # elif isinstance(value, torch.Tensor) or isinstance(value, Buffering):
        #     for hook in torch.nn.modules.module._global_buffer_registration_hooks.values():
        #         output = hook(self, name, value)
        #         if output is not None:
        #             value = output
        #     object.__getattribute__(self, '_contents')[name] = value
        else:
            object.__setattr__(self, name, value)
    
    @bufferingmethod
    @hidden
    def delete_attribute(self, name: str):
        try:
            object.__getattribute__(self, '_contents').pop(name)
        except KeyError:
            object.__delattr__(self, name)
    
    @bufferingmethod
    @hidden
    def get_content(self):
        return self
    
    @bufferingmethod
    @hidden
    def get_persistent(self):
        return object.__getattribute__(self, '__persistent')

    # def __copy__(self):
    #     from copy import copy
    #     instance = type(self).__new__(type(self))
    #     object.__setattr__(instance, '__persistent', object.__getattribute__(self, '__persistent'))
    #     for name, tensor in object.__getattribute__(self, '_contents').items():
    #         object.__getattribute__(instance, '_contents')[name] = copy(tensor)
    #     self_dict: dict = object.__getattribute__(self, '__dict__')
    #     try:
    #         self_dict.pop('_contents'), self_dict.pop('__persistent'), self_dict.pop('__class')
    #     except: pass
    #     for name, value in self_dict.items():
    #         type(self).__setattr__(instance, name, copy(value))
    #     return instance
    
    # def __deepcopy__(self, memodict={}):
    #     from copy import deepcopy
    #     instance = type(self).__new__(type(self))
    #     object.__setattr__(instance, '__persistent', object.__getattribute__(self, '__persistent'))
    #     for name, tensor in object.__getattribute__(self, '_contents').items():
    #         object.__getattribute__(instance, '_contents')[name] = deepcopy(tensor, memodict)
    #     self_dict: dict = object.__getattribute__(self, '__dict__')
    #     try:
    #         self_dict.pop('_contents'), self_dict.pop('__persistent'), self_dict.pop('__class')
    #     except: pass
    #     for name, value in self_dict.items():
    #         type(self).__setattr__(instance, name, deepcopy(value, memodict))
    #     return instance
    
    def __repr__(self):
        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = ''
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split('\n')
        child_lines = []
        for key, tensor in object.__getattribute__(self, '_contents').items():
            mod_str = repr(tensor)
            mod_str = _addindent(mod_str, 2)
            child_lines.append('(' + key + '): ' + mod_str)
        lines = extra_lines + child_lines

        main_str = type(self).__name__ + '('
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += '\n  ' + '\n  '.join(lines) + '\n'

        main_str += ')'
        return main_str
    
    def copy_(self, src, non_blocking: bool = False):
        contents = object.__getattribute__(self, '_contents')
        src_contents = object.__getattribute__(src, '_contents')
        for name in contents.keys():
            contents[name].copy_(src_contents[name], non_blocking)
        return self
    
    def to(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def cuda(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def ipu(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def xpu(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def cpu(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def type(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def float(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def double(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def half(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def bfloat16(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def detach(self, *args, **kwargs): return _named_forward_copy(self, *args, **kwargs)
    def is_floating_point(self): return False
    def is_complex(self): return False
    def to_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def cuda_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def ipu_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def xpu_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def cpu_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def type_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def float_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def double_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def half_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def bfloat16_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)
    def detach_(self, *args, **kwargs): return _named_forward_modify(self, *args, **kwargs)


class BufferDict(BufferObject):
    _contents: Mapping[str, Any]
    
    def __init__(self, buffers: Optional[Mapping[str, Any]] = None):
        if buffers is not None:
            self.update(buffers)
    
    @bufferingmethod
    @hidden
    def get_attr(self, name: str) -> Any:
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    @bufferingmethod
    @hidden
    def set_attribute(self, name: str, value):
        object.__setattr__(self, name, value)
    
    @bufferingmethod
    @hidden
    def delete_attribute(self, name: str):
        object.__delattr__(self, name)
    
    def __getitem__(self, key: str) -> Any:
        return object.__getattribute__(self, '_contents')[key]
    
    def __setitem__(self, key: str, value):
        # if value is None or isinstance(value, torch.Tensor) or isinstance(value, Buffering):
        #     for hook in torch.nn.modules.module._global_buffer_registration_hooks.values():
        #         output = hook(self, key, value)
        #         if output is not None:
        #             value = output
        #     object.__getattribute__(self, '_contents')[key] = value
        # else:
        #     raise TypeError(f"`{type(self).__name__}` instance does not accept instance from other than `{torch.Tensor.__name__}`, `{Buffering.__name__}`, and `NoneType`")
        object.__getattribute__(self, '_contents')[key] = value
    
    def __delitem__(self, key: str):
        object.__getattribute__(self, '_contents').pop(key)
    
    def __reversed__(self):
        return reversed(object.__getattribute__(self, '_contents'))
    
    def __len__(self):
        return len(object.__getattribute__(self, '_contents'))
    
    def __contains__(self, key: str):
        return object.__getattribute__(self, '_contents').__contains__(key)
    
    def __iter__(self):
        return iter(object.__getattribute__(self, '_contents'))
    
    def update(self, buffers: Mapping[str, Any]):
        # updating = OrderedDict()
        # for key, value in new_dict.items():
        #     if value is None or isinstance(value, torch.Tensor) or isinstance(value, Buffering):
        #     for hook in torch.nn.modules.module._global_buffer_registration_hooks.values():
        #         output = hook(self, key, value)
        #         if output is not None:
        #             value = output
        #         updating[key] = value
        #     else:
        #         raise TypeError(f"`{type(self).__name__}` instance does not accept instance from other than `{torch.Tensor.__name__}`, `{Buffering.__name__}`, and `NoneType`")
        object.__getattribute__(self, '_contents').update(buffers)
    
    def get(self, key: str):
        return object.__getattribute__(self, '_contents').get(key)
    
    def pop(self, key: str):
        return object.__getattribute__(self, '_contents').pop(key)
    
    def popitem(self):
        return object.__getattribute__(self, '_contents').popitem()
    
    def clear(self):
        object.__getattribute__(self, '_contents').clear()
    
    def copy(self):
        return self.__copy__()
    
    def setdefault(self, key: str, default: Any = None) -> Any:
        return object.__getattribute__(self, '_contents').setdefault(key, default)
    
    def items(self):
        return object.__getattribute__(self, '_contents').items()
    
    def keys(self):
        return object.__getattribute__(self, '_contents').keys()
    
    def values(self):
        return object.__getattribute__(self, '_contents').values()


class BufferList(BufferObject):
    _contents: Iterable[Any]
    
    def __init__(self, buffers: Optional[Iterable[Any]] = None):
        if buffers is not None:
            self += buffers
    
    @bufferingmethod
    @hidden
    def modify_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        contents = object.__getattribute__(self, '_contents')
        for i, content in enumerate(contents):
            contents[i] = func(content, *args, **kwargs)
        return self
    
    @bufferingmethod
    @hidden
    def copy_generically(self, func: Callable[[Any], Any], *args, **kwargs):
        instance = super().__new__(type(self))
        self_dict: dict = object.__getattribute__(self, '__dict__')
        for name, value in self_dict.items():
            type(self).__setattr__(instance, name, value)
        for i, content in enumerate(object.__getattribute__(self, '_contents')):
            object.__getattribute__(instance, '_contents')[i] = func(content, *args, **kwargs)
        return instance
    
    @bufferingmethod
    @hidden
    @staticmethod
    def tensors():
        return []
    
    @bufferingmethod
    @hidden
    def get_attr(self, name: str) -> Any:
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    @bufferingmethod
    @hidden
    def set_attribute(self, name: str, value):
        object.__setattr__(self, name, value)
    
    @bufferingmethod
    @hidden
    def delete_attribute(self, name: str):
        object.__delattr__(self, name)
    
    def __getitem__(self, index: int) -> Any:
        return object.__getattribute__(self, '_contents')[index]
    
    def __setitem__(self, index: int, value):
        # if value is None or isinstance(value, torch.Tensor) or isinstance(value, Buffering):
        #     for hook in torch.nn.modules.module._global_buffer_registration_hooks.values():
        #         output = hook(self, f'_{index}', value)
        #         if output is not None:
        #             value = output
        object.__getattribute__(self, '_contents')[index] = value
    
    def __delitem__(self, index: int):
        object.__getattribute__(self, '_contents').pop(index)
    
    def __reversed__(self):
        return reversed(object.__getattribute__(self, '_contents'))
    
    def __len__(self):
        return len(object.__getattribute__(self, '_contents'))
    
    def __contains__(self, value):
        return object.__getattribute__(self, '_contents').__contains__(value)
    
    def __iter__(self):
        return iter(object.__getattribute__(self, '_contents'))
    
    def count(self, value: Any) -> int:
        return object.__getattribute__(self, '_contents').count(value)
    
    def pop(self, index: int):
        return object.__getattribute__(self, '_contents').pop(index)
    
    def clear(self):
        object.__getattribute__(self, '_contents').clear()
    
    def copy(self):
        return self.__copy__()
    
    def __iadd__(self, other: Iterable[Any]):
        object.__getattribute__(self, '_contents').__iadd__(other)
        return self

    def __add__(self, other: Iterable[Any]):
        combined = type(self)(object.__getattribute__(self, '_contents') + other)
        object.__setattr__(combined, '__persistent', object.__getattribute__(self, '__persistent'))
        return combined

    def insert(self, index: int, buffer: Any):
        # if buffer is None or isinstance(buffer, torch.Tensor) or isinstance(buffer, Buffering):
        #     for hook in torch.nn.modules.module._global_buffer_registration_hooks.values():
        #         output = hook(self, f'_{index}', buffer)
        #         if output is not None:
        #             buffer = output
        object.__getattribute__(self, '_contents').insert(index, buffer)

    def append(self, buffer: Any):
        object.__getattribute__(self, '_contents').append(buffer)
        return self

    def extend(self, buffers: Iterable[Any]):
        object.__getattribute__(self, '_contents').extend(buffers)
        return self

    # def __copy__(self):
    #     from copy import copy
    #     instance = type(self).__new__(type(self))
    #     object.__setattr__(instance, '__persistent', object.__getattribute__(self, '__persistent'))
    #     for value in object.__getattribute__(self, '_contents'):
    #         object.__getattribute__(instance, '_contents').append(copy(value))
    #     self_dict: dict = object.__getattribute__(self, '__dict__')
    #     try:
    #         self_dict.pop('_contents'), self_dict.pop('__persistent'), self_dict.pop('__class')
    #     except: pass
    #     for name, value in self_dict.items():
    #         type(self).__setattr__(instance, name, copy(value))
    #     return instance
    
    # def __deepcopy__(self, memodict={}):
    #     from copy import deepcopy
    #     instance = type(self).__new__(type(self))
    #     object.__setattr__(instance, '__persistent', object.__getattribute__(self, '__persistent'))
    #     for value in object.__getattribute__(self, '_contents'):
    #         object.__getattribute__(instance, '_contents').append(deepcopy(value, memodict))
    #     self_dict: dict = object.__getattribute__(self, '__dict__')
    #     try:
    #         self_dict.pop('_contents'), self_dict.pop('__persistent'), self_dict.pop('__class')
    #     except: pass
    #     for name, value in self_dict.items():
    #         type(self).__setattr__(instance, name, deepcopy(value, memodict))
    #     return instance

    def __repr__(self):
        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = ''
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split('\n')
        child_lines = []
        for i, tensor in enumerate(object.__getattribute__(self, '_contents')):
            key = f'{i}'
            mod_str = repr(tensor)
            mod_str = _addindent(mod_str, 2)
            child_lines.append('(' + key + '): ' + mod_str)
        lines = extra_lines + child_lines

        main_str = type(self).__name__ + '('
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += '\n  ' + '\n  '.join(lines) + '\n'

        main_str += ')'
        return main_str


from typing import overload, Optional, Dict, Union, Mapping, Any, Iterable
from _collections_abc import dict_items, dict_keys, dict_values
import torch
from torch import device, dtype
from torch import nn

def bufferingmethod(funcobj): ...
class Buffering: ...
class Buffer(torch.Tensor, Buffering):
    @overload
    def __init__(self, tensor: Optional[torch.Tensor], persistent: bool = True): ...
class BufferObject(Buffering):
    _contents: Dict[str, Union[torch.Tensor, Buffering, None]]
    @property
    def persistent(self) -> bool: ...
    @persistent.setter
    def persistent(self, value): ...
    def with_persistent(self, value) -> BufferObject: ...
    def register_to(self, module: nn.Module, name: str): ...
    def copy_(self, src: BufferObject, non_blocking: bool = False) -> BufferObject:  ...
    @overload
    def to(self, device: Optional[Union[int, device]] = ..., dtype: Optional[Union[dtype, str]] = ..., non_blocking: bool = ...) -> BufferObject: ...
    @overload
    def to(self, dtype: Union[dtype, str], non_blocking: bool = ...) -> BufferObject: ...
    @overload
    def to(self, tensor: torch.Tensor, non_blocking: bool = ...) -> BufferObject: ...
    def cuda(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def ipu(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def xpu(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def cpu(self) -> BufferObject: ...
    def type(self, dst_type: Union[dtype, str]) -> BufferObject: ...
    def float(self) -> BufferObject: ...
    def double(self) -> BufferObject: ...
    def half(self) -> BufferObject: ...
    def bfloat16(self) -> BufferObject: ...
    def detach(self, *args, **kwargs) -> BufferObject: ...
    def is_floating_point(self, *args, **kwargs) -> bool: ...
    def is_complex(self, *args, **kwargs) -> bool: ...
    @overload
    def to_(self, device: Optional[Union[int, device]] = ..., dtype: Optional[Union[dtype, str]] = ..., non_blocking: bool = ...) -> BufferObject: ...
    @overload
    def to_(self, dtype: Union[dtype, str], non_blocking: bool = ...) -> BufferObject: ...
    @overload
    def to_(self, tensor: torch.Tensor, non_blocking: bool = ...) -> BufferObject: ...
    def cuda_(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def ipu_(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def xpu_(self, device: Optional[Union[int, device]] = None) -> BufferObject: ...
    def cpu_(self) -> BufferObject: ...
    def type_(self, dst_type: Union[dtype, str]) -> BufferObject: ...
    def float_(self) -> BufferObject: ...
    def double_(self) -> BufferObject: ...
    def half_(self) -> BufferObject: ...
    def bfloat16_(self) -> BufferObject: ...
    def detach_(self, *args, **kwargs) -> BufferObject: ...
class BufferDict(BufferObject):
    _contents: Mapping[str, Any]
    def update(self, buffers: Mapping[str, Any]): ...
    def get(self, key: str) -> Any: ...
    def pop(self, key: str) -> Any: ...
    def popitem(self) -> tuple[str, Any]: ...
    def clear(self): ...
    def copy(self) -> BufferDict: ...
    def setdefault(self, key: str, default: Any = None) -> Any: ...
    def items(self) -> dict_items[str, Any]: ...
    def keys(self) -> dict_keys[str, Any]: ...
    def values(self) -> dict_values[str, Any]: ...
class BufferList(BufferObject):
    _contents: Iterable[Any]
    def count(self, value: Any) -> int: ...
    def pop(self, index: int) -> Any: ...
    def clear(self): ...
    def copy(self) -> BufferList: ...
    def insert(self, index: int, buffer: Any): ...
    def append(self, buffer: Any): ...
    def extend(self, buffers: Iterable[Any]): ...


from typing import Callable, Iterable, Sequence, Optional, Union, Tuple
from enum import Enum, auto
import torch
from torch.types import _dtype
import numpy as np
from array_like_generic import *

def apply_from_dim(func: Callable, tensor: torch.Tensor, dim = 0, otypes: Iterable[_dtype] = None):
    '''
    Select a dim of an `torch.Tensor` and apply a function.
    '''
    slices = (slice(None, None, None),) * dim
    if len(otypes) > 1:
        return tuple(torch.as_tensor(item, dtype=otypes[i], device=tensor.device) for i, item in enumerate(zip(*[func(tensor[slices + (i,)]) for i in range(tensor.shape[dim])])))
    elif len(otypes) == 1:
        return torch.as_tensor([func(tensor[slices + (i,)]) for i in range(tensor.shape[dim])], otypes[0], device=tensor.device)
    else:
        [func(tensor[slices + (i,)]) for i in range(tensor.shape[dim])]

def min_dims(tensor: torch.Tensor, dims: Sequence[int], keepdim=False, *, out: Union[torch.Tensor, Tuple[torch.Tensor, ...], Sequence[torch.Tensor], None] = None):
    from .private import _min_max_dims
    return _min_max_dims(torch.min, tensor, dims, keepdim, out=out)

def max_dims(tensor: torch.Tensor, dims: Sequence[int], keepdim=False, *, out: Union[torch.Tensor, Tuple[torch.Tensor, ...], Sequence[torch.Tensor], None] = None):
    from .private import _min_max_dims
    return _min_max_dims(torch.max, tensor, dims, keepdim, out=out)

def map_range(
    tensor: torch.Tensor,
    interval: Sequence[int] = (0, 1),
    dim: Union[int, Sequence[int], None] = None,
    dtype: torch.dtype = None,
    scalar_default: ScalarDefault = ScalarDefault.max,
    eps: float = 1e-6
):
    min_value: torch.Tensor = min_dims(tensor, dims=dim, keepdim=True)
    max_value: torch.Tensor = max_dims(tensor, dims=dim, keepdim=True)
    max_min_difference = max_value - min_value
    max_min_equal_mask = max_min_difference == 0
    max_min_difference.masked_fill_(max_min_equal_mask, 1)
    tensor = (tensor - min_value).masked_fill(max_min_equal_mask, torch.tensor(scalar_default_value(scalar_default, eps)).to(tensor.dtype))
    return (tensor / max_min_difference * (interval[1] - interval[0]) + interval[0]).to(dtype)

def map_ranges(
    tensor: torch.Tensor,
    intervals: Sequence[Sequence[int]] = [(0, 1)],
    dim: Union[int, Sequence[int], None] = None,
    dtype: torch.dtype = None,
    scalar_default: ScalarDefault = ScalarDefault.max,
    eps: float = 1e-6
):
    min_value: torch.Tensor = min_dims(tensor, dims=dim, keepdim=True)
    max_value: torch.Tensor = max_dims(tensor, dims=dim, keepdim=True)
    max_min_difference = max_value - min_value
    max_min_equal_mask = max_min_difference == 0
    max_min_difference.masked_fill_(max_min_equal_mask, 1)
    normed = (tensor - min_value) / max_min_difference
    results = np.empty(len(intervals), dtype=torch.Tensor)
    for i, interval in enumerate(intervals):
        normed = normed.masked_fill(max_min_equal_mask, torch.tensor(scalar_default_value(scalar_default, eps)).to(tensor.dtype))
        results[i] = (normed * (interval[1] - interval[0]) + interval[0]).to(dtype)
    return results

def gamma(input: torch.Tensor, *, out: torch.Tensor | None = None):
    return torch.exp(torch.lgamma(input, out=out), out=out)

def gamma_div(left: torch.Tensor, right: torch.Tensor, *, out: torch.Tensor | None = None):
    return torch.exp(torch.lgamma(left, out=out) - torch.lgamma(right, out=out), out=out)

def recur_lgamma(n: torch.Tensor, base: torch.Tensor):
    out = torch.lgamma(base + n) - torch.lgamma(base)
    gamma_sign = torch.empty_like(base, dtype=torch.int64)
    gamma_sign[base > 0] = 1
    negative_one_exponent = torch.ceil(-base).to(device=base.device, dtype=int)
    negative_one_exponent[negative_one_exponent > n] = torch.broadcast_to(n, negative_one_exponent.shape)[negative_one_exponent > n]
    negative_base_mask = base < 0
    even_mod_exponent = negative_one_exponent % 2
    gamma_sign[negative_base_mask & (even_mod_exponent == 0)] = 1
    gamma_sign[negative_base_mask & (even_mod_exponent != 0)] = -1
    return out, gamma_sign

def arith_gamma_prod(arith_term: torch.Tensor, arith_base: torch.Tensor, ratio_base: torch.Tensor):
    arith_lgamma, gamma_sign = recur_lgamma(arith_term, arith_base)
    ratio_base_sign = torch.sign(ratio_base)
    ratio_base_sign[(ratio_base_sign == -1) & (arith_term % 2 == 0)] = 1
    return torch.exp(arith_term * torch.log(torch.abs(ratio_base)) + arith_lgamma) * gamma_sign * ratio_base_sign

def linspace(start, stop, num, dtype: Optional[_dtype] = None):
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = torch.as_tensor((stop - start) / (num - 1).to(dtype=dtype))
    index = torch.arange(num).to(common_difference.device)
    return start + common_difference * index.view([*index.shape] + [1] * len(common_difference.shape))

def linspace_at(index, start, stop, num, dtype: Optional[_dtype] = None):
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = torch.as_tensor((stop - start) / (num - 1).to(dtype=dtype))
    index = torch.as_tensor(index)
    return start + common_difference * index.view([*index.shape] + [1] * len(common_difference.shape))

def linspace_cover(index, start, stop, num, dtype: Optional[_dtype] = None):
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = (stop - start) / (num - 1).to(dtype=dtype)
    index = torch.as_tensor(index)
    return start + common_difference * index.view([*index.shape] + [1] * (len(common_difference.shape) - len(index.shape)))

def linspace_cumprod_at(index, start, stop, num, dtype: Optional[_dtype] = None):
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = (stop - start) / (num - 1).to(dtype=dtype)
    index = torch.as_tensor(index)
    result_index_prefix = (slice(None),) * len(index.shape)
    n = index + 1
    n_shape = n.shape
    result = torch.zeros((*n.shape, *common_difference.shape), device=common_difference.device, dtype=dtype)
    zero_common_difference_mask = common_difference == 0
    left_required_shape, right_required_shape = broadcast_required_shape(start[zero_common_difference_mask].shape, n_shape)
    n = n.view(right_required_shape)
    result[combine_mask_index(result_index_prefix, zero_common_difference_mask)] = torch.pow(start[zero_common_difference_mask].view(left_required_shape), n)
    sequence_with_zero_mask = torch.zeros_like(zero_common_difference_mask, device=common_difference.device, dtype=bool)
    nonzero_common_difference_mask = common_difference != 0
    sequence_with_zero_mask[nonzero_common_difference_mask] = ((start[nonzero_common_difference_mask] % common_difference[nonzero_common_difference_mask] == 0)
                                                               & (torch.sign(start[nonzero_common_difference_mask]) != torch.sign(common_difference[nonzero_common_difference_mask])))
    del nonzero_common_difference_mask
    result[combine_mask_index(result_index_prefix, sequence_with_zero_mask)] = 0
    first_divided_mask = torch.logical_not(sequence_with_zero_mask | zero_common_difference_mask)
    del zero_common_difference_mask
    del sequence_with_zero_mask
    first_divided = torch.full(common_difference.shape, torch.nan, device=common_difference.device, dtype=torch.float64)
    first_divided[first_divided_mask] = start[first_divided_mask] / common_difference[first_divided_mask].to(torch.float64)
    neg_int_arith_base_mask = ((torch.abs(torch.round(first_divided) - first_divided) < torch.pow(10, torch.floor(torch.log10(first_divided)) + 9))
                               & (torch.sign(first_divided) != 1))
    non_neg_int_arith_base_mask = torch.logical_not(neg_int_arith_base_mask) & first_divided_mask
    neg_int_arith_base_mask &= first_divided_mask
    if torch.any(non_neg_int_arith_base_mask == True):
        left_required_shape, right_required_shape = broadcast_required_shape(common_difference[non_neg_int_arith_base_mask].shape, n_shape)
        n = n.view(right_required_shape)
        result[combine_mask_index(result_index_prefix, non_neg_int_arith_base_mask)] = arith_gamma_prod(n, first_divided[non_neg_int_arith_base_mask].view(left_required_shape), common_difference[non_neg_int_arith_base_mask].view(left_required_shape)).to(result.dtype)
    del non_neg_int_arith_base_mask
    if torch.any(neg_int_arith_base_mask == True):
        left_required_shape, right_required_shape = broadcast_required_shape(common_difference[neg_int_arith_base_mask].shape, n_shape)
        n = n.view(right_required_shape)
        result[combine_mask_index(result_index_prefix, neg_int_arith_base_mask)] = arith_gamma_prod(n, -first_divided[neg_int_arith_base_mask].view(left_required_shape) - n + 1, -common_difference[neg_int_arith_base_mask].view(left_required_shape)).to(result.dtype)
    return result

def linspace_cumprod_cover(index, start, stop, num, dtype: Optional[_dtype] = None):
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = (stop - start) / (num - 1).to(dtype=dtype)
    index = torch.as_tensor(index)
    n = index + 1
    n_shape = n.shape
    result = torch.zeros(common_difference.shape, device=common_difference.device, dtype=dtype)
    zero_common_difference_mask = common_difference == 0
    n_required_shape = broadcast_large_dims(common_difference.shape, n_shape)
    n = n.view(n_required_shape)
    result[zero_common_difference_mask] = start[zero_common_difference_mask].to(result.dtype)
    result = torch.pow(result, n)
    sequence_with_zero_mask = torch.zeros_like(zero_common_difference_mask, device=common_difference.device, dtype=bool)
    nonzero_common_difference_mask = common_difference != 0
    sequence_with_zero_mask[nonzero_common_difference_mask] = ((start[nonzero_common_difference_mask] % common_difference[nonzero_common_difference_mask] == 0)
                                                               & (torch.sign(start[nonzero_common_difference_mask]) != torch.sign(common_difference[nonzero_common_difference_mask])))
    del nonzero_common_difference_mask
    result[sequence_with_zero_mask] = 0
    first_divided_mask = torch.logical_not(sequence_with_zero_mask | zero_common_difference_mask)
    del zero_common_difference_mask
    del sequence_with_zero_mask
    first_divided = torch.full(common_difference.shape, torch.nan, device=common_difference.device, dtype=torch.float64)
    first_divided[first_divided_mask] = start[first_divided_mask] / common_difference[first_divided_mask].to(torch.float64)
    neg_int_arith_base_mask = ((torch.abs(torch.round(first_divided) - first_divided) < torch.pow(10, torch.floor(torch.log10(first_divided)) + 9))
                               & (torch.sign(first_divided) != 1))
    non_neg_int_arith_base_mask = torch.logical_not(neg_int_arith_base_mask) & first_divided_mask
    neg_int_arith_base_mask &= first_divided_mask
    if torch.any(non_neg_int_arith_base_mask == True):
        result[non_neg_int_arith_base_mask] = arith_gamma_prod(n, first_divided, common_difference)[non_neg_int_arith_base_mask].to(result.dtype)
    del non_neg_int_arith_base_mask
    if torch.any(neg_int_arith_base_mask == True):
        result[neg_int_arith_base_mask] = arith_gamma_prod(n, -first_divided - n + 1, -common_difference)[neg_int_arith_base_mask].to(result.dtype)
    return result

def invert(tensor: torch.Tensor):
    shape_length = len(tensor.shape)
    if shape_length > 2:
        dims = (shape_length - 2, shape_length - 1)
    else:
        dims = (shape_length - 1,)
    min_values = min_dims(tensor, dims=dims, keepdim=True)
    max_values = max_dims(tensor, dims=dims, keepdim=True)
    return max_values - tensor + min_values

def buffer(tensor: Optional[torch.Tensor], persistent: bool = True) -> torch.Tensor:
    from .nn import Buffer
    return Buffer(tensor, persistent)

def advanced_indexing(shape, indices, dim):
    from .private import _limited_dim_indexing
    len_shape = len(shape)
    indices_shape = indices.shape
    left_args = _limited_dim_indexing(shape[:dim], len_shape)
    right_args = _limited_dim_indexing(shape[dim + 1:], len_shape, dim + 1)
    return *left_args, indices.view((*indices_shape,) + (1,) * (len_shape - len(indices_shape))), *right_args

class DimsGrowthDirection(Enum):
    leading = auto()
    trailing = auto()

def grow_dims(tensor: torch.Tensor, expected_dims_num: int, direction: DimsGrowthDirection = DimsGrowthDirection.leading):
    tensor_shape = tensor.shape
    remaining_dims_num = expected_dims_num - len(tensor_shape)
    if direction == DimsGrowthDirection.leading:
        return tensor.view((*((1,) * remaining_dims_num), *tensor_shape))
    else:
        return tensor.view((*tensor_shape, *((1,) * remaining_dims_num)))

def shift(tensor: torch.Tensor, shift: Sequence[int], fill_value = 0):
    empty = torch.full_like(tensor, fill_value=fill_value)
    shift_slices, container_slices = [np.empty((len(shift),), dtype=object) for _ in range(2)]
    for i, current_shift in enumerate(shift):
        if current_shift > 0:
            shift_slices[i] = slice(current_shift, None, None)
            container_slices[i] = slice(None, -current_shift, None)
        elif current_shift < 0:
            shift_slices[i] = slice(None, current_shift, None)
            container_slices[i] = slice(-current_shift, None, None)
        else:
            shift_slices[i] = slice(None, None, None)
            container_slices[i] = slice(None, None, None)
    shift_slices = (..., *shift_slices)
    container_slices = (..., *container_slices)
    empty[container_slices] = tensor[shift_slices]
    return empty

# def view_gather(input: torch.Tensor, dim: int, index: torch.Tensor):
#     return input[advanced_indexing(input.shape, index, dim)]

# def map_range(tensor: torch.Tensor, range: Sequence[int] = (0, 1), dim = None, dtype=None, scalar_default = ScalarDefault.max):
#     assert range[0] < range[1]
#     min_value = min_dims(tensor, dims=dim, keepdim=True)
#     max_value = max_dims(tensor, dims=dim, keepdim=True)
#     result = torch.empty(tensor.shape, dtype=dtype, device=tensor.device)
#     unrangable_index = torch.where(max_value == min_value)[0]
#     if len(unrangable_index) > 0:
#         result[unrangable_index] = scalar_default_value(scalar_default, range)
#     rangable_index = torch.where(max_value != min_value)[0]
#     if len(rangable_index) > 0:
#         rangable_min = min_value[rangable_index]
#         result[rangable_index] = ((tensor[rangable_index] - rangable_min) / (max_value[rangable_index] - rangable_min) * (range[1] - range[0]) + range[0]).to(result.dtype)
#     return result

# def map_ranges(tensor: torch.Tensor, ranges: Sequence[int] = [(0, 1)], dim = None, dtype=torch.float32, scalar_default = ScalarDefault.max):
#     min_value = min_dims(tensor, dims=dim, keepdim=True)
#     max_value = max_dims(tensor, dims=dim, keepdim=True)
#     unrangable_index = torch.where(max_value == min_value)[0]
#     rangable_index = torch.where(max_value != min_value)[0]
#     results = np.empty(len(ranges), dtype=object)
#     rangable_min = min_value[rangable_index]
#     normed = ((tensor[rangable_index] - rangable_min) / (max_value[rangable_index] - rangable_min)).to(dtype=dtype)
#     del rangable_min
#     for i, range in enumerate(ranges):
#         assert range[0] < range[1]
#         results[i] = torch.empty(tensor.shape, dtype=dtype, device=tensor.device)
#         if len(unrangable_index) > 0:
#             results[i][unrangable_index] = scalar_default_value(scalar_default, range)
#         if len(rangable_index) > 0:
#             results[i][rangable_index] = (normed * (range[1] - range[0]) + range[0]).to(dtype=dtype)
#     return results

# torch.apply_from_dim = apply_from_dim
# torch.map_range = map_range
# torch.map_ranges = map_ranges
# torch.linspace_at = linspace_at
# torch.linspace_cover = linspace_cover
# torch.linspace_cumprod_at = linspace_cumprod_at
# torch.linspace_cumprod_cover = linspace_cumprod_cover
# torch.invert = invert
# torch.buffer = buffer


import torch
import torch.nn.functional as F
from typing import Union, Sequence


def convert_to_floating_image(image: torch.Tensor):
    if image.is_floating_point():
        return image
    else:
        dtype_max = torch.iinfo(image.dtype).max
        image = image / dtype_max
        return image


def sobel_edges(image: torch.Tensor):
    """Returns a tensor holding Sobel edge maps.
    
    :param image: Image tensor with shape [batch_size, channels, height, width], expected a floating point type.
    
    :return: Tensor holding edge maps for each channel. Returns a tensor with shape
             [batch_size, channels, height, width, 2] where the last dimension holds [sobel_edge_y, sobel_edge_x].
    """
    # Sobel Filters
    kernels = torch.tensor([[[-1, -2, -1], [0, 0, 0], [1, 2, 1]], [[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]], dtype=image.dtype, device=image.device)
    kernels = kernels[:, None, :, :].repeat(image.size(1), 1, 1, 1)
    
    padded_image = F.pad(image, [1, 1, 1, 1], mode='reflect')
    output = F.conv2d(padded_image, kernels, groups=image.size(1)).view(image.size(0), image.size(1), -1, image.size(2), image.size(3)).permute(0, 1, 3, 4, 2)
    return output


def shift_image(image: torch.Tensor, shift: Union[int, Sequence[int]], fill_value = 0):
    from ..torchflint import shift as shift_
    from torch.nn.modules.utils import _pair
    if not (shift is int or len(shift) == 2):
        raise TypeError(f"'shift' should be either an 'int' or a pair of 'int'.")
    shift = _pair(shift)
    return shift_(image, shift, fill_value)


from typing import Optional
import torch
from torch.nn import Module, DataParallel
from torch.nn.parallel import DistributedDataParallel
from .buffer import bufferingmethod, Buffering, Buffer, BufferObject, BufferDict, BufferList


def refine_model(model: Module):
    if isinstance(model, DataParallel) or isinstance(model, DistributedDataParallel):
        return model.module
    else:
        return model